"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.matchObject = exports.matchArray = exports.matchAny = void 0;
/**
 * Verifies that the provided `obj` value matches the provided `toMatch` value
 *
 * @param obj The value that should match the `toMatch` property
 * @param toMatch The value that the `obj` property should match
 */
function matchAny(obj, toMatch) {
    if (typeof obj !== typeof toMatch) {
        return false;
    }
    if (typeof obj === 'object' && obj !== null && toMatch !== null) {
        if (Array.isArray(obj) !== Array.isArray(toMatch)) {
            return false;
        }
        if (Array.isArray(obj)) {
            return matchArray(obj, toMatch);
        }
        return matchObject(obj, toMatch);
    }
    return obj === toMatch;
}
exports.matchAny = matchAny;
/**
 * Verifies that each values of `match` array are contained in `array` array
 *
 * @param array The array that should contain every values of `match` array
 * @param match The array that should be contained in `array`
 */
function matchArray(array, match) {
    if (array.length < match.length) {
        return false;
    }
    const arrayCopy = [];
    for (let i = 0; i < array.length; i++) {
        arrayCopy[i] = array[i];
    }
    for (let i = 0; i < match.length; i++) {
        const toMatch = match[i];
        let found = false;
        for (let j = 0; j < arrayCopy.length; j++) {
            if (matchAny(arrayCopy[j], toMatch)) {
                // Remove the value from the arrayCopy so we don't match it twice
                // and this reduces the number of iterations we need to do over the arrayCopy
                arrayCopy.splice(j, 1);
                found = true;
                break;
            }
        }
        /**
           * If there is no value in the array that matches the value we want to match,
           * then the array doesn't match
           */
        if (!found) {
            return false;
        }
    }
    return true;
}
exports.matchArray = matchArray;
/**
 * Verifies that each properties of `match` object are contained in `obj` object
 *
 * @param obj The object that should contain every properties of `match` object
 * @param match The object that should be contained in `obj`
 */
function matchObject(obj, match) {
    /**
     * Why not using Object.keys()?
     *
     * Object.keys() forces us to iterate over all properties of the object to list them first, before we can iterate over them.
     *
     * In this case we might early exit if we find a property that doesn't match, this means
     * we might not test every properties.
     * So, to reduce the overhead we iterate over the properties of the match object one by one as we test them.
     * This is way faster than doing Object.keys() and then iterating over the result when Objects gets bigger.
     * On objects with ~20 properties, this is ~2x faster than doing Object.keys().
     */
    for (const key in match) {
        if (!matchAny(obj[key], match[key])) {
            return false;
        }
    }
    return true;
}
exports.matchObject = matchObject;
//# sourceMappingURL=ObjectMatcher.js.map